//
// Module: BadPixelCorrector
// Author: Richard Saxton
//
// Description: 
//
// A class to provide corrections for bad pixels which are taken from
// extensions in the event file
//
// Modifications: See ChangeLog
// 
// $Id: BadPixelCorrector.cc,v 1.4 2007/04/26 09:41:56 rsaxton Exp $

/* General */

/* SAS-specific */
#include "Tracer.h"
#include "CalServer.h"
#include "Dal.h"
#include "DataFilter.h"
#include "OalServer.h"
#include "EventAttributes.h"

/* Local */
#include "NumericalRange.h"
#include "ArfDataServer.h"
#include "BadPixelCorrector.h"
#include "BadPixPSFCorrector.h"
#include "PixelIterator.h"
#include "CompactTablePixelIteratorRaw.h"
#include "OffCCDPSFCorrector.h"

/* Constructor */
BadPixelCorrector::BadPixelCorrector(const string & badpixContainer, 
                   unsigned eefactor, PsfModel * psf, unsigned long flags)
  // :_eefactor(eefactor)
{
  PROC;
   
    /* Set the name for the temporary bad column file */
    bctempfile="temp_badcol_arfgen.del"; 

    /* Open the file containing the bad pixel values */

    if (dataSetServer->exists(badpixContainer)) {

     /* Create a temporary bad pixel file from all the bad columns
        and bad pixels contained in the original badpixcontainer */
      _bcset = createTempBadColFile(badpixContainer, bctempfile, 
                                                   flags);

      errstr << "Created semi-temporary bad column file - "<<bctempfile
             << Sas::message(Msg::App, Msg::Verbose);
 
      /* Set up an array of bad pixel correctors - one for each CCD */

      for (unsigned int ccd = 1; ccd <= 12; ccd++) {

        string bpstring = ccd < 10 ? "BADPIX0":"BADPIX";
        string table = bpstring+toString(ccd);

        errstr << "Checking for table: " << table
               << Sas::message(Msg::App, Msg::Verbose);

     /* If bad column or pixel data exist for this CCD, add it to the vector */
        if (_bcset->hasBlock(table)) {
           _bpcorr.push_back(new BadPixPSFCorrector(bctempfile+":"+table,
                                               eefactor, psf, ccd));
        }

      }
     /* Close the bad pixel sets */
      dataSetServer->close(_bcset);


    } 
    else { // No BadPixelFile specified or found

       errstr<<" Warning couldn't find dataset: "<<badpixContainer
             <<" - correcting for chip gaps but not for bad pixels"
             << Sas::warning("NoBadPixLocations");
    }

   /* Add an extra corrector for pixels which are outside the CCD boundaries */
    errstr<<"Adding OFF CCD corrector"
               << Sas::message(Msg::App, Msg::Verbose);
   
    _bpcorr.push_back(new OffCCDPSFCorrector(eefactor, psf));

    errstr<<"BadPix size now: "<<_bpcorr.size()
               << Sas::message(Msg::App, Msg::Verbose);
   
  ENDPROC;
}

/* Destructor */
BadPixelCorrector::~BadPixelCorrector()
{
  PROC;
   /* Attempt to close the temp bad column file. Force a remove
      if this fails */
     if ( dataSetServer->exists(bctempfile))
          dataSetServer->remove(bctempfile);

  ENDPROC;
}

/* Method: TotalFactor(energyvector)
         : Calculates total bad pixel effect folded through PSF etc. */

vector<double>
BadPixelCorrector::TotalFactor(const vector<double> & energyVector) const
{
  PROC;

   /* Define the corrections vector and initialise to zero */
    vector<double> corrFactor, tempvec;

    for (unsigned int i = 0; i < energyVector.size(); ++i) {
       corrFactor.push_back(0.0);
    }

  errstr<<"BadPix size: "<<_bpcorr.size()
               << Sas::message(Msg::App, Msg::Verbose);
   
   /* Filter the bad pixels to exclude those falling outside the 
      source region */
    for (unsigned int c= 0; c < _bpcorr.size(); c++) {
  errstr<<"BadPix filtering map: "<<c
               << Sas::message(Msg::App, Msg::Verbose);
   
        _bpcorr[c]->filterMap();
    }

   /* Add the correction factors together for each chip at each energy */
    for (unsigned int c= 0; c < _bpcorr.size(); c++) {
  errstr<<"BadPix calling Fraction: "<<c
               << Sas::message(Msg::App, Msg::Verbose);
   
         //corrFactor[i] += (*_bpcorr)[c]->Fraction(energyVector)[i];
        tempvec = _bpcorr[c]->Fraction(energyVector);

        for (unsigned int i = 0; i < energyVector.size(); ++i) {
            corrFactor[i] += tempvec[i];
        }
    }

  errstr<<"BadPix total 1:"<<corrFactor[0]<<" 5:"<<corrFactor[4]
               << Sas::message(Msg::App, Msg::Verbose);
   
  RETURN corrFactor;
}
  
int
BadPixelCorrector::totalPixels() const
{
  PROC;

    int total=0;

   /* Add up the discarded detector pixels from each element */ 
    for (unsigned int c= 0; c < _bpcorr.size(); c++) {

      errstr<<"BadPix - adding up pixels: "<<c
               << Sas::message(Msg::App, Msg::Verbose);
   
      total += _bpcorr[c]->totalPixels();
    
errstr<<"BPtot= "<<total
    << Sas::message(Msg::App, Msg::Verbose); 
    }
  
  RETURN total;
}

/* Method createTempBadColFile(badpixContainer, bctempfile, flags) 
      - makes a temp file containing bad column data in form of bad pixels */
DataSet * BadPixelCorrector::createTempBadColFile(const string & badpixContainer, const string & bctempfile, unsigned long flags) const
{
  PROC;

     /* Defs: */
      Table * bpixtab;
      int16 * orawxData;
      int16 * orawyData; int16 * otypeData; 
      int16 * oextentData; int16 * oflagData;
      unsigned numccds=12;
      unsigned long tableelements=0;                                                     
      ColumnData * rawxColData=0, * ccdColData=0, * offxColData=0;
      int8 * ccdData=0;
      int16 * rawxData=0, * offxData=0;
      int16 x1=0, x2=0, y1=0, y2=0;
      
     /* Open the input file */
      DataSet * bpset = dataSetServer->open( badpixContainer,
                                                 DataSet::Read );

     /* Create a temporary file for bad column data */
      DataSet * bcset = dataSetServer->open(bctempfile, DataSet::Temp);

     /* Use the selected event flagging to decide whether pixels next
        to certain types of bad pixels and columns and the CCD border
        or observation window should be excluded */
      vector<int16> extra_badpix;
      int16 extra_badcol=1, extra_edgeccd=1, extra_edgewindow=1;

     /* Set default for all bad pixels to yes. Only Bright, flickering and Dead 
        pixels are used so far (I think) so unset these as required */
      for (unsigned i=0; i<6; i++){extra_badpix.push_back(1);}  

     // Adjacent to HOT, flickering and dead  pixels
      if ( (flags & 1<<EventAttributes::CLOSE_TO_BRIGHTPIX)) 
                            extra_badpix[1] = 0;  // Bright pixel
      if ( (flags & 1<<EventAttributes::CLOSE_TO_FLICKERINGPIX)) 
                            extra_badpix[2] = 0;  // Flickering pixel
      if ( (flags & 1<<EventAttributes::CLOSE_TO_DEADPIX)) 
                            extra_badpix[3] = 0; // Dead pixel

     // Adjacent to bad columns
      if ( (flags & 1<<EventAttributes::CLOSE_TO_BADCOL)) 
                            extra_badcol = 0;    // Bad column

     // Edge of CCD or edge of observation window
      if ((flags & 1<<EventAttributes::CLOSE_TO_CCD_BORDER)) 
                            extra_edgeccd = 0;     // Edge of CCD
      if ((flags & 1<<EventAttributes::CLOSE_TO_CCD_WINDOW)) 
                            extra_edgewindow = 0;  // Edge of window

      errstr << "Flagging: "<<extra_badpix[0]<<" "<<extra_badpix[1]<<" "
             <<extra_badpix[2]<<" "<<extra_badpix[3]<<" "
             <<extra_badcol<<" "<<extra_edgeccd<<" "<<extra_edgewindow
               << Sas::message(Msg::App, Msg::Verbose);

     /* Find the range of CCD pixels which are not already excluded
        by chip gaps */
      int16 minx = static_cast<int16>(calServer.instrument()->minX())
                       + extra_edgeccd;
      int16 maxx = static_cast<int16>(calServer.instrument()->maxX())
                       - extra_edgeccd;
      int16 miny = static_cast<int16>(calServer.instrument()->minY())
                       + extra_edgeccd;
      int16 maxy = static_cast<int16>(calServer.instrument()->maxY())
                       - extra_edgeccd;

      errstr << "CCD extent "<<minx<<" "<<maxx<<" "<<miny<<" "<<maxy
               << Sas::message(Msg::App, Msg::Verbose);

     /* Process the bad column table if it exists */
      if (bpset->hasBlock("OFFSETS"))
      {
         Table * bctable = bpset->table("OFFSETS");
         tableelements = bctable->rows();

         errstr << "Found "<<tableelements<<" entries in OFFSETS table"
                << Sas::message(Msg::App, Msg::Verbose);

        /* If the table has some rows process them */
         if ( tableelements > 0 && bctable->hasColumn("RAWX") ) 
         {
          /* Map the input columns */
           rawxColData    = bctable->column("RAWX")->data();
           ccdColData     = bctable->column("CCDNR")->data();
           offxColData    = bctable->column("OFFSETX")->data();
           rawxData = rawxColData->int16Data();
           ccdData = ccdColData->int8Data();
           offxData = offxColData->int16Data();
         }
         else {
           tableelements=0;  // set to 0 to avoid processing if RAWX is missing
         }
      }

     /* Loop over each CCD */
      for (unsigned int ic=1; ic<=numccds; ic++) {

       /* Create a map to hold all the pixels in this CCD.
          Add one extra element to avoid the starting from 0 problem */
        SImage *twodvector = new SImage(maxx+1, maxy+1);
        SImage & ccdPixMap = *twodvector;
          
        string bpstring = ic < 10 ? "BADPIX0":"BADPIX";
        string tablename = bpstring+toString(ic);

        int16 counter = 0;

       /* Set pixels contained in the bad columns to 1 in the CCD map */
        for (unsigned int i=0; i<tableelements; i++) 
        {
            /* Set the x element to a local variable */
             int16 value = rawxData[i];
             int16 most = value + extra_badcol, least = value - extra_badcol;
             
            /* If the X offset = o then this is an offset row 
               rather than a column (MOS only) */
             bool columnNotRow = offxData[i] > 0 ? true : false;

            /* Process this record if it belongs to this CCD */
             if ( ccdData[i] == ic )
             {
                counter++;

              /* Find the affected pixels and set in the map.
                 Make sure not to overlap with chip edge exclusions */
                if (columnNotRow) {
                   x1 = max(minx, least);
                   x2 = min(maxx, most);
                   y1 = miny;
                   y2 = maxy;
               } else {
                   y1 = max(miny, least);
                   y2 = min(maxy, most);
                   x1 = minx;
                   x2 = maxx;
               }
               for (int16 ix=x1; ix<x2+1; ix++)
                  for (int16 iy=y1; iy<y2+1; iy++)
                     ccdPixMap[ix][iy]=1;
             }   
        } // End of loop over table entries

       /* Set pixels contained in bad pixel extensions to 1 in the CCD map */
        if (bpset->hasBlock(tablename))  // if badpix extn exists for this CCD
        {
          // Open the bad pixel extension as an iterator
           PixelIterator *ctpxitr = new CompactTablePixelIteratorRaw(
                              badpixContainer+":"+tablename, 
                              "RAWX","RAWY","TYPE","YEXTENT");

          // Increment the counter
           counter++;

          // Loop over each pixel
           FOREACH(ctpxitr)
           {
             PixelElement pixel = ctpxitr->currentItem();
                                                                                
            /* Calculate the limits of this pixel including the adjacent
               pixels if required */
             int16 ixc = static_cast<int16>(pixel.xCentre());
             int16 iyc = static_cast<int16>(pixel.yCentre());
             int16 least = ixc - extra_badpix[UNT(pixel.value())];
             int16 most = ixc + extra_badpix[UNT(pixel.value())];
             x1 = max(minx, least);
             x2 = min(maxx, most);
             least = iyc - extra_badpix[UNT(pixel.value())];
             most = iyc + extra_badpix[UNT(pixel.value())];
             y1 = max(miny, least);
             y2 = min(maxy, most);

             int16 ixc2 = min(maxx,max(minx, ixc));
             int16 iyc2 = min(maxy,max(miny, iyc));

            /* Set these pixels to one in the CCD map */
             for (int16 ix=x1; ix<x2+1; ix++)
                            { ccdPixMap[ix][iyc2]=1; }
             for (int16 iy=y1; iy<y2+1; iy++)
                            { ccdPixMap[ixc2][iy]=1; }

           } // end of badpixel iterator

           delete ctpxitr;

        } // end of badpixel block

        errstr << "Found "<<counter<<" entries for CCD "<<ic
               << Sas::message(Msg::App, Msg::Noisy);

       /* Create a table for this CCD if records exist */
        if (counter > 0) {

          /* Dont know number of rows at this stage - contract later */
           int fullsize=10000;
           bpixtab = bcset->addTable(tablename, fullsize);

          /* Map the output columns */
           Column * orawxColumn = bpixtab->addColumn("RAWX",
                                                       Column::Int16,
                                                      "Raw X value",
                                                      "pixel");

           Column * orawyColumn = bpixtab->addColumn("RAWY",
                                                       Column::Int16,
                                                      "Raw Y value",
                                                      "pixel");

           Column * otypeColumn = bpixtab->addColumn("TYPE",
                                                       Column::Int16,
                                                      "Bad pixel type",
                                                      "");

           Column * oextentColumn = bpixtab->addColumn("YEXTENT",
                                                       Column::Int16,
                                                      "Y direction extent",
                                                      "pixel");

           Column * oflagColumn = bpixtab->addColumn("BADFLAG",
                                                       Column::Int16,
                                                      "Bad pixel flag",
                                                      "");

           ColumnData * orawxColData = orawxColumn->data();
           orawxData = orawxColData->int16Data();

           ColumnData * orawyColData = orawyColumn->data();
           orawyData = orawyColData->int16Data();

           ColumnData * otypeColData = otypeColumn->data();
           otypeData = otypeColData->int16Data();

           ColumnData * oextentColData = oextentColumn->data();
           oextentData = oextentColData->int16Data();

           ColumnData * oflagColData = oflagColumn->data();
           oflagData = oflagColData->int16Data();

          /* Copy data to columns */
           int16 ystart=0;
           int16 cnt=-1;

           for (int16 ix=minx; ix<maxx+1; ix++)
           {
             bool start = false;

             for (int16 iy=miny; iy<maxy+1; iy++)
             {
               if (ccdPixMap[ix][iy]==1 && ! start ) 
               {
                  cnt++;
                  start = true;
                  ystart = iy;
                  orawxData[cnt]=ix;
                  orawyData[cnt]=iy;
                  otypeData[cnt]=1;
                  oflagData[cnt]=1;
               }
               if (ccdPixMap[ix][iy]==0 && start)
               {
                 oextentData[cnt]=(iy-ystart);
                 start = false;
               }
             }
             if (start) oextentData[cnt]=(maxy - ystart + 1);
           }
          /* Cut the table size down to the actual number of rows */
           bpixtab->deleteRows( cnt+1, (fullsize - cnt - 1) );

          /* Delete pointers to created table */
/* Comment this out because it seg faults for some reason
         errstr << "Removing columns"
               << Sas::message(Msg::App, Msg::Verbose);
     
           delete orawxColData;
           delete orawyColData;
           delete otypeColData;
           delete oextentColData;

           delete oflagColData;
         errstr << "Removed columns"
               << Sas::message(Msg::App, Msg::Verbose);
*/
     
         } // End of records exist block
       } // End of CCD loop

     /* Remove pointers to input data */
       delete rawxColData;
       delete ccdColData;
       delete offxColData;

     // } // End of if rows>0 block
        
     /* Uncomment this line to crash the program and 
                             keep the temporary offsets table */
     //dataSetServer->close(bcset);

     dataSetServer->close(bpset);

    /* Return the dataset pointer */
     RETURN bcset;
}
/* Method checkAdd(rawx, int16 value) - add value to vector if not
                                        already present */
// Not used any more
bool BadPixelCorrector::checkAdd(vector<int16> rawx2, int16 value,
bool ignoreEdgePixels) const
{
  PROC;

   /* Check if this value already exists in the vector */
     bool oktoAdd=true;

     vector<int16> rawx = rawx2;

     for (unsigned int i=0; i<rawx.size(); i++) 
     {
        if (rawx[i] == value) { oktoAdd=false; }
     }
        
   /* Also check that this value isn't off chip or next to a chip gap 
      which has already been ignored */

     int extra_edgeccd=0;
     if (ignoreEdgePixels) {
        extra_edgeccd=1;
     }

     if ( (value < INT(calServer.instrument()->minX()+extra_edgeccd)) ||
          (value > INT(calServer.instrument()->maxX()-extra_edgeccd)) )
     {
        oktoAdd=false;
     } 

     RETURN ( oktoAdd );
                        
  ENDPROC;
}


